open Import

val care_about_ite_branch : bool ref
(** Ignored -- kept for compatibility. *)

val about_ite_branch_ref : (Location.t -> unit) ref
(** Ignored -- kept for compatibility. *)

val about_ite_branch : Location.t -> unit
(** Ignored -- kept for compatibility. *)
