(** Magic numbers *)

val ast_intf_magic_number : string
(** Magic number for file holding an interface syntax tree *)

val ast_impl_magic_number : string
(** Magic number for file holding an implementation syntax tree *)
